/*  XMMS plugin for raw cdr files (based on orginal xmms wav plugin) 
 *  Copyright (C) 2001 Marek Blaszkowski <mb@kis.p.lodz.pl>, <mb@intercon.pl>
 *
 *  Wav plugin Copyright (C) 1998-1999  Peter Alm, Mikael Alm, Olle Hallnas, 
 *  Thomas Nilsson and 4Front Technologies
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 *  1.1: 2002-09-29 Herbert Straub <h.straub@aon.at>
 *  	1) cdr_init (): read from ~/.xmms/config the swap byte order paramter
 *  	2) configure_ok_cb (): store the swap byte order paramter 
 *  	   in ~/.xmms/config
 *	    3) is_our_file (): read also .raw files
 */

#include <glib.h>
#include <gtk/gtk.h>
#include <xmms/configfile.h>

#include "cdr.h"
#include "util.h"


#define FORMAT FMT_S16_LE   
/* I don't know why FMT_U16_BE doesn't work, so I choosed as above and
 * I swaped input data later before send it to output 
 */

#define rate 176400

InputPlugin cdr_ip =
{
	NULL,
	NULL,
	"Raw .cdr file player 1.1",
	cdr_init,
	cdr_about,
	cdr_conf,
	is_our_file,
	NULL,
	play_file,
	stop,
	cdr_pause,
	seek,
	NULL,
	get_time,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	get_song_info,
	NULL,
	NULL
};

CdrFile *cdr_file = NULL;
static GtkWidget *cdr_about_win=NULL;
static GtkWidget *cdr_config_win=NULL;
static pthread_t decode_thread;
static gboolean audio_error = FALSE;
static gboolean swapdata = FALSE;

InputPlugin *get_iplugin_info(void)
{
	return &cdr_ip;
}

static void cdr_init(void)
{
	ConfigFile *cfgfile;
	gchar *filename;

	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);
	cfgfile = xmms_cfg_open_file(filename);
	if (cfgfile)
	{
		xmms_cfg_read_boolean(cfgfile, "rawcdr", "swapByteOrder", &swapdata);
		xmms_cfg_free(cfgfile);
	}
	g_free(filename);
}

static int is_our_file(char *filename)
{
	gchar *ext;

	ext = strrchr(filename, '.');
	if (ext)
		if (!strcasecmp(ext, ".cdr")) return TRUE;
		if (!strcasecmp(ext, ".raw")) return TRUE;
	return FALSE;
}

static void swap(char *buf,int size)
{

int i;
char val1,val2;

for (i=0;i<size-1;i=i+2){ /* size of buf data */
		val1=buf[i];
		val2=buf[i+1];
		buf[i]=val2;
		buf[i+1]=val1;
		}
}

static void *play_loop(void *arg)
{
char data[2048 * 2]; /* buf must have even size */
int bytes, blk_size;
int actual_read;

blk_size = 512 * (16 / 8) * 2;
while (cdr_file->going)
	{
	 if(!cdr_file->eof)
	  {
		 bytes = blk_size;
		if (cdr_file->length - cdr_file->position < bytes)
                bytes = cdr_file->length - cdr_file->position;
		if (bytes > 0)
            {
                actual_read = fread(data, 1, bytes, cdr_file->file);
				if (!swapdata)swap(data,sizeof(data));
                if (actual_read == -1)cdr_file->eof = 1;
                else
				  { 
                    cdr_ip.add_vis_pcm(cdr_ip.output->written_time(),FORMAT,
                       2, bytes, data);
                    while(cdr_ip.output->buffer_free() < bytes && cdr_file->going && cdr_file->seek_to == -1)
                        xmms_usleep(10000);
                    if(cdr_file->going && cdr_file->seek_to == -1)
                        cdr_ip.output->write_audio(data, bytes);
                    cdr_file->position += actual_read;
                }
            }
            else
            {	
                cdr_file->eof = TRUE;
                xmms_usleep(10000);
            }
        }
        else
            xmms_usleep(10000);
        if (cdr_file->seek_to != -1)
        {
            cdr_file->position = cdr_file->seek_to * rate;
            fseek(cdr_file->file, cdr_file->position + cdr_file->data_offset, SEEK_SET);
            cdr_ip.output->flush(cdr_file->seek_to * 1000);
            cdr_file->seek_to = -1;
        }
	  }
		fclose(cdr_file->file);
		pthread_exit(NULL);
}

static void play_file(char *filename)
{
char *name, *temp;
struct stat s;
int status;

cdr_file=malloc(sizeof (CdrFile));
memset(cdr_file, 0, sizeof (CdrFile));
if ((cdr_file->file = fopen(filename, "rb")))
	{
		status=stat (filename, &s);
		if(status!=0){
					fclose(cdr_file->file);
            		free(cdr_file);
            		cdr_file = NULL;
            		return;	
					}	
			 cdr_file->data_offset = 0;		
			 cdr_file->length = s.st_size;
			 cdr_file->position = 0;
             cdr_file->going = 1; 
			if (cdr_ip.output->open_audio(FORMAT, 44100, 2) == 0){
            		audio_error = TRUE;
            		fclose(cdr_file->file);
            		free(cdr_file);
            		cdr_file = NULL;
            		return;
        			}
			
		temp = strrchr(filename, '/');
        if (!temp) temp = filename;
        		else temp++;
        name = malloc(strlen(temp) + 1);
        strcpy(name, temp);
        *strrchr(name, '.') = '\0';
        cdr_ip.set_info(name, 1000 * (cdr_file->length / rate), 8 * rate, 
				44100, 2);
        free(name);
        cdr_file->seek_to = -1;
		pthread_create(&decode_thread, NULL, play_loop, NULL);
	}
}

static void stop(void)
{
	if (cdr_file)
	{
		if (cdr_file->going)
		{
			cdr_file->going = 0;
			pthread_join(decode_thread, NULL);
			cdr_ip.output->close_audio();
			free(cdr_file);
			cdr_file = NULL;

		}
	}
}

static void cdr_pause(short p)
{
	cdr_ip.output->pause(p);
}

static void seek(int time)
{
	cdr_file->seek_to = time;

	cdr_file->eof = FALSE;

	while (cdr_file->seek_to != -1)
		xmms_usleep(10000);
}

static int get_time(void)
{
 if (audio_error)
        return -2;
    if (!cdr_file)
        return -1;
    if (!cdr_file->going || (cdr_file->eof && !cdr_ip.output->buffer_playing()))
        return -1;
    else
    {
        return cdr_ip.output->output_time();
    }
}

static void get_song_info(char *filename, char **title, int *length)
{
char *name, *temp;
unsigned long len;
CdrFile *cdr_file;
struct stat s;
int status;

cdr_file = malloc(sizeof (CdrFile));
memset(cdr_file, 0, sizeof (CdrFile));
if (!(cdr_file->file = fopen(filename, "rb"))) return;
status=stat (filename, &s);
if(status!=0){
             fclose(cdr_file->file);
             free(cdr_file);
             cdr_file = NULL;
             return;
             }
len = s.st_size;
(*length) = 1000 * (len / rate);

    temp = strrchr(filename, '/');
    if (!temp)
        temp = filename;
    else
        temp++;
    name = g_malloc(strlen(temp) + 1);
    strcpy(name, temp);
    *strrchr(name, '.') = '\0';
    (*title) = name;

    fclose(cdr_file->file);
    free(cdr_file);
    cdr_file = NULL;
}

void cdr_about(void)
{
GtkWidget *vbox, *frame_info,*frame_warn, *label_info,*label_warn,*box, *button;

#define WARN " The raw cdr (ripped from a CDROM) files doesn't \n \
contain any header. They have always 44100 \n \
samples/sec, 16 bit, 2 channels, big endian \n \
order (MSBL LSBL, MSBR LSBR). This file \n \
type is detected only by a .cdr extention. \n \
If xmms find a *.cdr file and this file \n \
is not a .cdr raw audio file, then you'll \n \
get noise and/or it may damage your hi-fi \n \
speakers if volume is loud."

if (cdr_about_win) return;
cdr_about_win = gtk_window_new(GTK_WINDOW_DIALOG);
gtk_signal_connect(GTK_OBJECT(cdr_about_win), "destroy", GTK_SIGNAL_FUNC
				(gtk_widget_destroyed), &cdr_about_win);
gtk_window_set_title(GTK_WINDOW(cdr_about_win), "Raw cdr input");
gtk_window_set_policy(GTK_WINDOW(cdr_about_win), FALSE, FALSE, FALSE);
gtk_window_set_position(GTK_WINDOW(cdr_about_win), GTK_WIN_POS_MOUSE);
gtk_container_border_width(GTK_CONTAINER(cdr_about_win), 10);
vbox = gtk_vbox_new(FALSE, 10);
gtk_container_add(GTK_CONTAINER(cdr_about_win), vbox);
frame_info = gtk_frame_new("Version 1.1");
gtk_box_pack_start(GTK_BOX(vbox), frame_info, FALSE, FALSE, 0);
label_info = gtk_label_new("Raw CDROM audio files player.\nMarek Blaszkowski\nEmail: mb@kis.p.lodz.pl, mb@intercon.pl\nV1.1 Herbert Straub\nmail: h.straub@aon.at\n");
gtk_container_add(GTK_CONTAINER(frame_info), label_info);
frame_warn = gtk_frame_new("Warning");
gtk_box_pack_start(GTK_BOX(vbox), frame_warn, FALSE, FALSE, 0);
label_warn = gtk_label_new(WARN);
gtk_label_set_justify(GTK_LABEL(label_warn), GTK_JUSTIFY_LEFT);
gtk_container_add(GTK_CONTAINER(frame_warn), label_warn);
box = gtk_hbutton_box_new();
gtk_button_box_set_spacing(GTK_BUTTON_BOX(box), 5);
gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 0);
button = gtk_button_new_with_label("Ok");
gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC
					(gtk_widget_destroy), GTK_OBJECT(cdr_about_win));
GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
gtk_widget_grab_default(button);
gtk_widget_show(button);
gtk_widget_show(box);
gtk_widget_show(frame_info);
gtk_widget_show(frame_warn);
gtk_widget_show(label_info);
gtk_widget_show(label_warn);
gtk_widget_show(vbox);
gtk_widget_show(cdr_about_win);
}

static void set_swap (GtkWidget *widget, gpointer data)
{
if (GTK_TOGGLE_BUTTON (widget)->active)  swapdata=TRUE;
	else swapdata=FALSE;
}

static void configure_ok_cb(gpointer data)
{
	gchar *filename;
	ConfigFile *cfgfile;

	filename = g_strconcat(g_get_home_dir(), "/.xmms/config", NULL);

	cfgfile = xmms_cfg_open_file(filename);
	if (!cfgfile)
		cfgfile = xmms_cfg_new();
	xmms_cfg_write_boolean(cfgfile, "rawcdr", "swapByteOrder", swapdata);
	xmms_cfg_write_file(cfgfile, filename);
	xmms_cfg_free(cfgfile);
	g_free(filename);
	gtk_widget_destroy(cdr_config_win);
}


static void cdr_conf(void)
{
GtkWidget *vbox, *frame,*box, *button,*tgbutton;

if (cdr_config_win) return;
cdr_config_win = gtk_window_new(GTK_WINDOW_DIALOG);
gtk_signal_connect(GTK_OBJECT(cdr_config_win), "destroy", GTK_SIGNAL_FUNC
                (gtk_widget_destroyed), &cdr_config_win);
gtk_window_set_title(GTK_WINDOW(cdr_config_win),"Raw cdr input");
gtk_window_set_policy(GTK_WINDOW(cdr_config_win), FALSE, FALSE, FALSE);
gtk_window_set_position(GTK_WINDOW(cdr_config_win), GTK_WIN_POS_MOUSE);
gtk_container_border_width(GTK_CONTAINER(cdr_config_win), 10);
vbox = gtk_vbox_new(FALSE, 10);
gtk_container_add(GTK_CONTAINER(cdr_config_win), vbox);
frame = gtk_frame_new("Config");
gtk_box_pack_start(GTK_BOX(vbox), frame, FALSE, FALSE, 0);
tgbutton=gtk_check_button_new_with_label("swap byte order on input file");
gtk_signal_connect (GTK_OBJECT (tgbutton), "toggled",
				GTK_SIGNAL_FUNC (set_swap),NULL);
if(swapdata)
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tgbutton),1);
gtk_container_add(GTK_CONTAINER(frame), tgbutton);
box = gtk_hbutton_box_new();
gtk_button_box_set_spacing(GTK_BUTTON_BOX(box), 5);
gtk_box_pack_start(GTK_BOX(vbox), box, FALSE, FALSE, 0);
button = gtk_button_new_with_label("Ok");
gtk_signal_connect(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC(configure_ok_cb), NULL);
/*
gtk_signal_connect_object(GTK_OBJECT(button), "clicked", GTK_SIGNAL_FUNC
                    (gtk_widget_destroy), GTK_OBJECT(cdr_config_win));
*/
GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
gtk_box_pack_start(GTK_BOX(box), button, TRUE, TRUE, 0);
gtk_widget_grab_default(button);
gtk_widget_show(button);
gtk_widget_show(box);
gtk_widget_show(frame);
gtk_widget_show(tgbutton);
gtk_widget_show(vbox);
gtk_widget_show(cdr_config_win);
}
